import os
import json
import base64

import requests

from binstar_client.utils import compute_hash
from binstar_client.requests_ext import stream_multipart

__version__ = '0.2.0'

# from poster.encode import multipart_encode
# from poster.streaminghttp import register_openers
# import urllib2
# register_openers()

class BinstarError(Exception):
    pass

class Unauthorized(BinstarError):
    pass

class NotFound(IndexError, BinstarError):
    pass

def jencode(payload):
    return base64.b64encode(json.dumps(payload))

class Binstar():
    '''
    An object that represents interfaces with the binstar.org restful API.

    :param token: a token generated by Binstar.authenticate or None for
                  an anonymous user.
    '''

    def __init__(self, token=None, domain='https://api.binstar.org'):
        self.session = requests.Session()
        self.session.headers['x-binstar-api-version'] = __version__
        self.token = token
        if token:
            self.session.headers.update({'Authorization': 'token %s' % (token)})

        self.domain = domain

    def authenticate(self, username, password, application, application_url=None, scopes=['package'],):
        '''
        Use basic authentication to create an authentication token using the interface below.
        With this technique, a username and password need not be stored permanently, and the user can
        revoke access at any time.

        :param username: The users name
        :param password: The users password
        :param application: The application that is requesting access
        :param application_url: The application's home page
        :param scopes: Scopes let you specify exactly what type of access you need. Scopes limit access for the tokens.
        '''

        url = '%s/authentications' % (self.domain)
        payload = {"scopes": scopes, "note": application, "note_url": application_url}
        data = base64.b64encode(json.dumps(payload))
        res = self.session.post(url, auth=(username, password), data=data, verify=True)
        self._check_response(res)
        res = res.json()
        token = res['token']
        self.session.headers.update({'Authorization': 'token %s' % (token)})
        return token

    def remove_authentication(self, token):
        url = '%s/authentications' % (self.domain)
        res = self.session.delete(url, verify=True)
        self._check_response(res)

    def _check_response(self, res, allowed=[200]):
        if not res.status_code in allowed:
            try:
                data = res.json()
            except:
                msg = 'Undefined server error (status code: %s)' % (res.status_code,)
            else:
                msg = data.get('error', 'Undefined server error (status code: %s)' %(res.status_code,))
            ErrCls = BinstarError
            if res.status_code == 401:
                ErrCls = Unauthorized
            elif res.status_code == 404:
                ErrCls = NotFound
            raise ErrCls(msg, res.status_code)

    def user(self, login=None):
        '''
        Get user infomration.

        :param login: (optional) the login name of the user or None. If login is None
                      this method will return the information of the authenticated user.
        '''
        if login:
            url = '%s/user/%s' % (self.domain, login)
        else:
            url = '%s/user' % (self.domain)

        res = self.session.get(url, verify=True)
        self._check_response(res)

        return res.json()

    def user_packages(self, login=None):
        '''
        Returns a list of packages for a given user

        :param login: (optional) the login name of the user or None. If login is None
                      this method will return the packages for the authenticated user.

        '''
        if login:
            url = '%s/packages/%s' % (self.domain, login)
        else:
            url = '%s/packages' % (self.domain)

        res = self.session.get(url, verify=True)
        self._check_response(res)

        return res.json()

    def package(self, login, package_name):
        '''
        Get infomration about a specific package

        :param login: the login of the package owner
        :param package_name: the name of the package
        '''
        url = '%s/package/%s/%s' % (self.domain, login, package_name)
        res = self.session.get(url, verify=True)
        self._check_response(res)

        return res.json()

    def all_packages(self, modified_after=None):
        '''
        '''
        url = '%s/package_listing' % (self.domain)
        data = {'modified_after':modified_after or ''}
        res = self.session.get(url, data=data, verify=True)
        self._check_response(res)
        return res.json()


    def add_package(self, login, package_name,
                    summary=None,
                    license=None,
                    public=True,
                    license_url=None,
                    attrs=None,
                    host_publicly=None):
        '''
        Add a new package to a users account

        :param login: the login of the package owner
        :param package_name: the name of the package to be created
        :param package_type: A type identifyer for the package (eg. 'pypi' or 'conda', etc.)
        :param summary: A short summary about the package
        :param license: the name of the package license
        :param license_url: the url of the package license
        :param public: if true then the package will be hosted publicly
        :param attrs: A dictionary of extra attributes for this package
        :param host_publicly: TODO: describe
        '''
        url = '%s/package/%s/%s' % (self.domain, login, package_name)

        attrs = attrs or {}
        attrs['summary'] = summary
        attrs['license'] = {'name':license, 'url':license_url}

        payload = dict(public=public,
                       public_attrs=attrs or {},
                       host_publicly=host_publicly)

        data = jencode(payload)
        res = self.session.post(url, verify=True, data=data)
        self._check_response(res)
        return res.json()

    def release(self, login, package_name, version):
        '''
        Get information about a specific release

        :param login: the login of the package owner
        :param package_name: the name of the package
        :param version: the name of the package
        '''
        url = '%s/release/%s/%s/%s' % (self.domain, login, package_name, version)
        res = self.session.get(url, verify=True)
        self._check_response(res)
        return res.json()

    def add_release(self, login, package_name, version, requirements, announce, description):
        '''
        Add a new release to a package.

        :param login: the login of the package owner
        :param package_name: the name of the package
        :param version: the version string of the release
        :param requirements: A dict of requirements TODO: describe
        :param announce: An announcement that will be posted to all package watchers
        :param description: A long description about the package
        '''

        url = '%s/release/%s/%s/%s' % (self.domain, login, package_name, version)

        payload = {'requirements':requirements, 'announce':announce, 'description':description}
        data = jencode(payload)
        res = self.session.post(url, data=data, verify=True)
        self._check_response(res)
        return res.json()

    def remove_dist(self, login, package_name, release, basename=None, _id=None):

        if basename:
            url = '%s/dist/%s/%s/%s/%s' % (self.domain, login, package_name, release, basename)
        elif _id:
            url = '%s/dist/%s/%s/%s/-/%s' % (self.domain, login, package_name, release, _id)
        else:
            raise TypeError("method remove_dist expects either 'basename' or '_id' arguments")

        print 'url', url
        res = self.session.delete(url, verify=True)
        self._check_response(res)
        return res.json()


    def download(self, login, package_name, release, basename, md5=None):
        '''
        Dowload a package distribution

        :param login: the login of the package owner
        :param package_name: the name of the package
        :param version: the version string of the release
        :param basename: the basename of the distribution to download
        :param md5: (optional) an md5 hash of the download if given and the package has not changed
                    None will be returned

        :returns: a file like object or None
        '''

        url = '%s/download/%s/%s/%s/%s' % (self.domain, login, package_name, release, basename)
        if md5:
            headers = {'ETag':md5, }
        else:
            headers = {}

        res = self.session.get(url, verify=True, headers=headers, allow_redirects=False)
        self._check_response(res, allowed=[302, 304])

        if res.status_code == 304:
            return None
        elif res.status_code == 302:
            res2 = requests.get(res.headers['location'], stream=True, verify=True)
            return res2.raw


    def upload(self, login, package_name, release, basename, fd, distribution_type, 
               description='', md5=None, size=None, attrs=None, callback=None):
        '''
        Upload a new distribution to a package release.

        :param login: the login of the package owner
        :param package_name: the name of the package
        :param version: the version string of the release
        :param basename: the basename of the distribution to download
        :param fd: a file like object to upload
        :param description: (optional) a short description about the file
        :param attrs: any extra attributes about the file (eg. build=1, pyversion='2.7', os='osx')

        '''
        url = '%s/stage/%s/%s/%s/%s' % (self.domain, login, package_name, release, basename)
        if attrs is None:
            attrs = {}
        if not isinstance(attrs, dict):
            raise TypeError('argument attrs must be a dictionary')

        payload = dict(distribution_type=distribution_type, description=description, attrs=attrs)
        data = jencode(payload)
        res = self.session.post(url, data=data, verify=True)
        self._check_response(res)
        obj = res.json()

        s3url = obj['s3_url']
        s3data = obj['s3form_data']

        if md5 is None:
            _hexmd5, b64md5, size = compute_hash(fd, size=size)
        elif size is None:
            spos = fd.tell()
            fd.seek(0, os.SEEK_END)
            size = fd.tell() - spos
            fd.seek(spos)

        s3data['Content-Length'] = size
        s3data['Content-MD5'] = b64md5

        data_stream, headers = stream_multipart(s3data, files={'file':(basename, fd)},
                                                callback=callback)

        s3res = requests.post(s3url, data=data_stream, verify=True, timeout=10*60*60, headers=headers)

        if s3res.status_code != 201:
            print s3res.text
            print
            print
            raise BinstarError('Error uploading to s3', s3res.status_code)

        url = '%s/commit/%s/%s/%s/%s' % (self.domain, login, package_name, release, basename)
        payload = dict(dist_id=obj['dist_id'])
        data = jencode(payload)
        res = self.session.post(url, data=data, verify=True)
        self._check_response(res)

        return res.json()


    def groups(self, owner=None):
        if owner:
            url = '%s/groups/%s' % (self.domain, owner)
        else:
            url = '%s/groups' % (self.domain, )

        res = self.session.get(url, verify=True)
        self._check_response(res)

        return res.json()

    def group(self, owner, group_name):
        url = '%s/group/%s/%s' % (self.domain, owner, group_name)
        res = self.session.get(url, verify=True)
        self._check_response(res)
        return res.json()

    def group_members(self, org, name):
        url = '%s/group/%s/%s/members' % (self.domain, org, name)
        res = self.session.get(url, verify=True)
        self._check_response(res)

        return res.json()

    def is_group_member(self, org, name, member):
        url = '%s/group/%s/%s/members/%s' % (self.domain, org, name,member)
        res = self.session.get(url, verify=True)
        self._check_response(res, [204, 404])
        return res.status_code == 204

    def add_group_member(self, org, name, member):
        url = '%s/group/%s/%s/members/%s' % (self.domain, org, name,member)
        res = self.session.put(url, verify=True)
        self._check_response(res, [204])
        return

    def remove_group_member(self, org, name, member):
        url = '%s/group/%s/%s/members/%s' % (self.domain, org, name,member)
        res = self.session.delete(url, verify=True)
        self._check_response(res, [204])
        return

    def remove_group_package(self, org, name, package):
        url = '%s/group/%s/%s/packages/%s' % (self.domain, org, name, package)
        res = self.session.delete(url, verify=True)
        self._check_response(res, [204])
        return

    def group_packages(self, org, name):
        url = '%s/group/%s/%s/packages' % (self.domain, org, name)
        res = self.session.get(url, verify=True)
        self._check_response(res, [200])
        return res.json()

    def add_group_package(self, org, name, package):
        url = '%s/group/%s/%s/packages/%s' % (self.domain, org, name, package)
        res = self.session.put(url, verify=True)
        self._check_response(res, [204])
        return

    def add_group(self, org, name, perms='read'):
        url = '%s/group/%s/%s' % (self.domain, org, name)

        payload = dict(perms=perms)
        data = jencode(payload)

        res = self.session.post(url, data=data, verify=True)
        self._check_response(res, [204])

        return




from binstar_client.utils import get_binstar, parse_specs
from argparse import FileType, ArgumentError
from os.path import basename
import tarfile
import json
from warnings import warn
from pprint import pprint
import re
from binstar_client.pprintb import package_list, user_list

def group_spec(spec):
    '''
    Spec
    '''
    pat = re.compile('^(?P<org>[a-z][a-z0-9_-]+)(/:(?P<group_name>[a-z][a-z0-9_-]+)(/(?P<member>[a-z][a-z0-9_-]+))?)?$')
    mat = pat.match(spec)
    if mat is None:
        raise ArgumentError('dfasdf')
    
    return mat.groupdict()

def main(args):
    
    binstar = get_binstar()
#    pprint(binstar.groups())
    
    if args.action == 'add':
        result = binstar.add_group(args.spec['org'], args.spec['group_name'], args.perms)
        pprint(result)
    elif args.action == 'show':
        if args.spec['group_name']:
            result = binstar.group(args.spec['org'], args.spec['group_name'])
            pprint(result)
        else:
            result = binstar.groups(args.spec['org'])
            pprint(result)
            
    elif args.action == 'members':
        if not args.spec['group_name']:
            raise ArgumentError('must specify group_name in spec')
        result = binstar.group_members(args.spec['org'], args.spec['group_name'])
        user_list(result, args.verbose)
    elif args.action == 'add_member':
        if not args.spec['group_name']:
            raise ArgumentError('must specify group_name in spec')
        if not args.spec['member']:
            raise ArgumentError('must specify group_name in spec')
        binstar.add_group_member(args.spec['org'], args.spec['group_name'], args.spec['member'])
    elif args.action == 'remove_member':
        if not args.spec['group_name']:
            raise ArgumentError('must specify group_name in spec')
        if not args.spec['member']:
            raise ArgumentError('must specify group_name in spec')
        binstar.remove_group_member(args.spec['org'], args.spec['group_name'], args.spec['member'])
    elif args.action == 'packages':
        if not args.spec['group_name']:
            raise ArgumentError('must specify group_name in spec')
        result = binstar.group_packages(args.spec['org'], args.spec['group_name'])
        package_list(result, args.verbose)
            
    elif args.action == 'add_package':
        if not args.spec['group_name']:
            raise ArgumentError('must specify group_name in spec')
        if not args.spec['member']:
            raise ArgumentError('must specify group_name in spec')
        binstar.add_group_package(args.spec['org'], args.spec['group_name'], args.spec['member'])
    elif args.action == 'remove_package':
        if not args.spec['group_name']:
            raise ArgumentError('must specify group_name in spec')
        if not args.spec['member']:
            raise ArgumentError('must specify group_name in spec')
        binstar.remove_group_package(args.spec['org'], args.spec['group_name'], args.spec['member'])
    else:
        raise NotImplementedError(args.action)

def add_parser(subparsers):
    
    parser = subparsers.add_parser('groups',
                                    help='Manage Groups',
                                    description=__doc__)

    parser.add_argument('action', help='asdf',
                        choices=['add', 'show', 'members', 'add_member', 'remove_member',
                                 'packages', 'add_package','remove_package'], nargs='?')
    parser.add_argument('spec', help=group_spec.__doc__, nargs='?', type=group_spec)
    parser.add_argument('--perms', help='group permissions', choices=['read', 'write', 'admin'], default='read')
    

    parser.set_defaults(main=main)


from binstar_client.utils import parse_specs, get_binstar
import tarfile
import json
from warnings import warn
from binstar_client import BinstarError, NotFound
from os.path import exists
import sys
import time
import yaml
from os.path import basename
from email.parser import Parser

def detect_yaml_attrs(filename):
    tar = tarfile.open(filename)
    obj = tar.extractfile('info/recipe/meta.yaml')
    attrs = yaml.load(obj)
    try:
        description = attrs['about']['home']
    except KeyError:
        description = None
    try:
        license = attrs['about']['license']
    except KeyError:
        license = None

    return description, license 

def detect_pypi_attrs(filename):
    
    with tarfile.open(filename) as tf:
        pkg_info = next(name for name in tf.getnames() if name.endswith('/PKG-INFO'))
        fd = tf.extractfile(pkg_info)
        attrs = dict(Parser().parse(fd).items())
        
    name = attrs.pop('Name')
    version = attrs.pop('Version')
    summary = attrs.pop('Summary')
    description = attrs.pop('Description')
    license = attrs.pop('License')
    attrs = {'dist':'sdist'}
    
    return name, version, attrs, summary, description, license

def detect_conda_attrs(filename):
    
    tar = tarfile.open(filename)
    obj = tar.extractfile('info/index.json')
    attrs = json.loads(obj.read())
    
    description, license = detect_yaml_attrs(filename)
    return attrs['name'], attrs['version'], attrs, description, description, license

detectors = {'conda':detect_conda_attrs,
             'pypi': detect_pypi_attrs,
             }


def detect_package_type(filename):
    
    if filename.endswith('.tar.bz2'): #Could be a conda package
        try:
            with tarfile.open(filename) as tf:
                tf.getmember('info/index.json')
        except KeyError:
            pass
        else:
            return 'conda'
    
    if filename.endswith('.tar.gz'): #Could be a setuptools sdist
        with tarfile.open(filename) as tf:
            if any(name.endswith('/PKG-INFO') for name in tf.getnames()):
                return 'pypi' 
    
    raise BinstarError('Could not autodetect the package type of file %s' % filename) 

def bool_input(prompt, default=True):
        while 1:
            inpt = raw_input('%s [Y|n]: ' % prompt)
            if inpt.lower() in ['y', 'yes'] and not default:
                return True
            elif inpt.lower() in ['', 'n', 'no'] and not default:
                return False
            elif inpt.lower() in ['', 'y', 'yes']:
                return True
            elif inpt.lower() in ['n', 'no']:
                return False
            else:
                print 'please enter yes or no'

def create_package(binstar,username, package_name, summary, license):
    binstar.add_package(username, package_name,
                        summary,
                        license)

def create_release(binstar, username, package_name, version, description, announce=None):
    binstar.add_release(username, package_name, version, [],
                        announce, description)


def create_package_interactive(binstar, username, package_name):
    
    print '\nThe package %s/%s does not exist' % (username, package_name)
    if not bool_input('Would you lke to create it now?'):
        print 'goodbbye'
        raise SystemExit(-1)
    
    summary = raw_input('Enter a short description of the package\nsummary: ')
    license = raw_input('Enter the name of the license (default:BSD)\nlicense: ')
    license_url = raw_input('Enter the url of the license (optional)\nlicense url: ')
    public = bool_input('\nDo you want to make this package public?')
    
    binstar.add_package(username, package_name,
                    summary,
                    license,
                    license_url,
                    public)

def create_release_interactive(binstar, username, package_name, version):
    
    print '\nThe release %s/%s/%s does not exist' % (username, package_name, version)
    if not bool_input('Would you like to create it now?'):
        print 'good-bye'
        raise SystemExit(-1)

    description = raw_input('Enter a short description of the release:\n')    
    print("\nAnnouncements are emailed to your package followers.")
    make_announcement = bool_input('Would you like to make an announcement to the package followers?', False)
    if make_announcement:
        announce = raw_input('Markdown Announcement:\n')
    else: 
        announce = ''
    
    binstar.add_release(username, package_name, version, [], 
                        announce, description)


def upload_print_callback():
    start_time = time.time()
    def callback(curr, total):
        curr_time = time.time()
        time_delta = curr_time - start_time
    
        remain = total - curr
        if curr and remain:
            eta =  1.0 * time_delta / curr * remain / 60.0
        else:
            eta = 0 
    
        curr_kb = curr//1024
        total_kb = total//1024
        perc = 100.0 * curr / total if total else 0
    
        msg = '\r uploaded %(curr_kb)i of %(total_kb)iKb: %(perc).2f%% ETA: %(eta).1f minutes'
        print msg % locals(),
        sys.stdout.flush()
        if curr == total:
            print

def main(args):
    
    binstar = get_binstar()
    
    if args.user:
        username = args.user
    else:
        user = binstar.user()
        username = user ['login']

    uploaded_packages = [] 

    for filename in args.files:

        if not exists(filename):
            raise BinstarError('file %s does not exist' %(filename)) 
    
        if args.package_type:
            package_type = args.package_type
        else:
            print 'detecting package type ...', 
            sys.stdout.flush()
            package_type = detect_package_type(filename)
            print package_type
        
        get_attrs = detectors[package_type]
    
        if args.metadata:
            attrs = json.loads(args.metadata)
            package_name = args.package
            version = args.version
        else:
            print 'extracting package attributes for upload ...',
            sys.stdout.flush()
            package_name, version, attrs, summary, description, license = get_attrs(filename)
            print 'done'

        if args.package:
            package_name = args.package

        if args.version:
            version = args.version

        try:
            binstar.package(username, package_name)
        except NotFound:
            if args.mode == 'interactive':
                create_package_interactive(binstar, username, package_name) 
            else:
                create_package(binstar, username, package_name, summary, license)   

        try:
            binstar.release(username, package_name, version)
        except NotFound:
            if args.mode == 'interactive':
                create_release_interactive(binstar, username, package_name, version)
            else:
                create_release(binstar, username, package_name, version, description)

        basefilename = basename(filename)
    
        with open(filename) as fd:
            print '\nUploading file %s/%s/%s/%s ... ' % (username, package_name, version, basefilename)
            sys.stdout.flush()
        
            binstar.upload(username, package_name, version, basefilename, fd, package_type, args.description, attrs=attrs, 
                       callback=upload_print_callback())

            uploaded_packages.append(package_name)


    print("\n\nUpload(s) Complete\n")
    for package in uploaded_packages:        
        print("Package located at:\nhttps://binstar.org/%s/%s\n" % (username, package))
    

def add_parser(subparsers):
    
    parser = subparsers.add_parser('upload',
                                      help='Upload a file to binstar',
                                      description=__doc__)
    
    parser.add_argument('files', nargs='*', help='Distributions to upload', default=[])

    parser.add_argument('-u', '--user', help='User account, defaults to the current user')
    parser.add_argument('-p', '--package', help='Defaults to the packge name in the uploaded file')
    parser.add_argument('-v', '--version', help='Defaults to the packge version in the uploaded file')
    parser.add_argument('-t', '--package-type', help='Set the package type, defaults to autodetect')
    parser.add_argument('-d','--description', help='description of the file(s)')
    parser.add_argument('-m','--metadata', help='json encoded metadata default is to autodetect')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-i', '--interactive', action='store_const', help='Run an interactive prompt if any packages are missing', 
                        dest='mode', const='interactive')
    group.add_argument('-f', '--fail', help='Fail if a package or release does not exist (default)', 
                                        action='store_const', dest='mode', const='fail' )
    
    parser.set_defaults(main=main)
    


'''
Print the information of the current user
'''
from binstar_client import Unauthorized
from binstar_client.utils import get_binstar

def main(args):
    binstar = get_binstar()
    
    try:
        user = binstar.user()
    except Unauthorized:
        return 'Anonymous User'
    
    for key_value in user.items():
        print  '%s: %s' % key_value

def add_parser(subparsers):
    subparser = subparsers.add_parser('whoami', 
                                      help='Print the information of the current user', 
                                      description=__doc__)
    
    subparser.set_defaults(main=main)


'''
Show information about an object:

example::

    * binstar show continuumio
    * binstar show continuumio/python
    * binstar show continuumio/python/2.7.5

    binstar show sean/meta/1.2.0/meta.tar.gz
    
     
'''
from binstar_client.utils import get_binstar, PackageSpec, parse_specs
from argparse import FileType, RawTextHelpFormatter
def main(args):
    
    binstar = get_binstar()
    
    spec = args.spec
    if spec._basename:
        print 'file'
    elif args.spec._version:
        print 'version', spec.version
        release = binstar.release(spec.user, spec.package, spec.version)
        for dist in release['distributions']:
            print '   + %(basename)s' % dist, ' platform:%(platform)-10s arch:%(arch)-10s' % dist['attrs']
        print 
        print '%(description)s' % release['public_attrs']
        
    elif args.spec._package:
        package = binstar.package(spec.user, spec.package)
        package['permission'] = 'public' if package['public'] else 'private'
        print '- %(name)s: [%(permission)s] %(summary)s'  % package
        for release in package['releases']:
            print '   + %(version)s' % release
            
    elif args.spec._user:
        print binstar.user(spec.user)
        for package in binstar.user_packages(spec.user):
            package['permission'] = 'public' if package['public'] else 'private'
            print '   + %(name)25s: [%(permission)s] %(summary)s'  % package
    else:
        print args.spec

def add_parser(subparsers):
    
    parser = subparsers.add_parser('show',
                                      help='Show information about an object',
                                      description=__doc__, formatter_class=RawTextHelpFormatter)
    
    parser.add_argument('spec', help='Package written as <user>[/<package>[/<version>[/<filename>]]]', type=parse_specs)
    
    
    
    parser.set_defaults(main=main)


'''
Authenticate a user
'''
from keyring import get_keyring
import getpass
from binstar_client import Binstar
from binstar_client.utils import get_config

def main(args):
    
    config = get_config()
    kr = get_keyring()
    token = kr.get_password('binstar-token', getpass.getuser())
    url = config.get('url', 'https://api.binstar.org')
    bs = Binstar(token, domain=url,)
    
    username = raw_input('Username: ')
    password = getpass.getpass()

    token = bs.authenticate(username, password, 'Binstar-Cli', url, ['packages'])
    kr.set_password('binstar-token', getpass.getuser(), token)
    print 'login successful'

def add_parser(subparsers):
    subparser = subparsers.add_parser('login', 
                                      help='Authenticate a user', 
                                      description=__doc__)
    
    subparser.set_defaults(main=main)


'''
Binstar command line utility
'''
from argparse import ArgumentParser
from binstar_client.commands import sub_commands
from binstar_client import BinstarError
import sys

def main():
    
    
    parser = ArgumentParser(description=__doc__)
    subparsers = parser.add_subparsers(help='commands')
    for command in sub_commands():
        command.add_parser(subparsers)
    
    parser.add_argument('--verbose', action='store_true')
    args = parser.parse_args()
    
    try:
        return args.main(args)
    except BinstarError as err:
        sys.stderr.write('%s: %s\n' %(type(err).__name__, err.args[0]))
        raise SystemExit(-1)


'''
Log out from binstar
'''

from keyring import get_keyring
import getpass
from binstar_client import Binstar
from binstar_client.utils import get_config

def main(args):
    
    config = get_config()
    kr = get_keyring()
    token = kr.get_password('binstar-token', getpass.getuser())
    url = config.get('url', 'https://api.binstar.org')
    bs = Binstar(token, domain=url)
    bs.remove_authentication(token)
    kr.delete_password('binstar-token', getpass.getuser())
    print("logout successful")

def add_parser(subparsers):
    subparser = subparsers.add_parser('logout', 
                                      help='Log out from binstar', 
                                      description=__doc__)
    
    subparser.set_defaults(main=main)


'''
Authenticate a user
'''
from binstar_client.utils import get_config, set_config, SITE_CONFIG,\
    USER_CONFIG

def main(args):
    config = get_config()
    
    if args.show:
        print 'Site Config:', SITE_CONFIG
        print get_config(user=False)
        print 'User Config:', USER_CONFIG
        print get_config(site=False)
    
    if args.get:
        print config[args.get]
        return
    
    if args.files:
        print 'User Config:', USER_CONFIG 
        print 'Site Config:', SITE_CONFIG
    
    config = get_config(args.user, not args.user)
    
    for key, value in args.set:
        config[key] = args.type(value)
    
    for key in args.remove:
        if key in config:
            del config[key]
    
    set_config(config, args.user)
    

def add_parser(subparsers):
    parser = subparsers.add_parser('config', 
                                      help='Authenticate a user', 
                                      description=__doc__)
    
    parser.add_argument('--type', type=eval, default=str)
    parser.add_argument('--set', nargs=2, action='append', default=[])
    parser.add_argument('--get')
    parser.add_argument('--remove', action='append', default=[])
    parser.add_argument('--show', action='store_true', default=False)
    parser.add_argument('-u', '--user', action='store_true', dest='user',default=True)
    parser.add_argument('-s', '--site', action='store_false', dest='user')
    parser.add_argument('-f','--files', action='store_true')
    
    
    parser.set_defaults(main=main)


'''
Add a new package to your account.
'''
from binstar_client.utils import get_binstar
import json
from pprint import pprint
from binstar_client.pprintb import package_list
def main(args):
    
    binstar = get_binstar()
    
    if '/' in args.spec: 
        user, package = args.spec.split('/', 1)
    else:
        user = args.spec
        package = None
        
    if args.attrs:
        with open(args.attrs) as fd:
            attrs = json.load(fd)
    else:
        attrs = {}
    
    if args.action == 'add':
        binstar.add_package(user, package, args.type,
                            args.summary, args.license, args.license_url,
                            args.public, attrs)
    elif args.action == 'show':
        print binstar.package(user, package)
    elif args.action == 'list':
        package_list(binstar.user_packages(user), args.verbose)
    else:
        raise NotImplementedError(args.action)


def add_parser(subparsers):
    
    parser = subparsers.add_parser('packages',
                                      help='Add a package',
                                      description=__doc__)
    
    parser.add_argument('action', help='Package name written as either <user>/<package> or just <package>',
                        choices=['add', 'list', 'show'])
    parser.add_argument('spec', help='Package name written as either <user>/<package> or just <package>')
    parser.add_argument('-t', '--type', help='package type (conda, pypi, etc)')
    parser.add_argument('-s', '--summary', help='Summary')
    parser.add_argument('-l', '--license', help='license name default:$(default)s', default='BSD')
    parser.add_argument('--license-url', help='license url', default='http://opensource.org/licenses/BSD-3-Clause')
    parser.add_argument('--public', help='Allow anyone to view and download this package',
                        action='store_true', dest='public', default=True)
    parser.add_argument('--private', help='You choose who can view has access to this package',
                        action='store_false', dest='public')
    parser.add_argument('--attrs', help='a json file containing any extra attributes you want this package to have')
    
    
    parser.set_defaults(main=main)


'''
Add a new package to your account.
'''
from binstar_client.utils import get_binstar
from argparse import FileType
def main(args):
    
    binstar = get_binstar()
    
    user, package, version = args.spec.split('/', 2)
    
    if args.announce:
        announce = args.announce.read()
    else:
        announce = ''
    if args.description:
        description = args.description.read()
    else:
        description = ''
    
    if args.action == 'add':    
        binstar.add_release(user, package, version, {}, announce, description)
    elif args.action == 'show':
        release = binstar.release(user, package, version)
        print release
    else:
        raise NotImplementedError(args.action)


def add_parser(subparsers):
    
    parser = subparsers.add_parser('release',
                                      help='Add a release',
                                      description=__doc__)
    
    parser.add_argument('action', help='Adde remove or update an existing release',
                        choices=['add', 'remove', 'update', 'show'])
    parser.add_argument('spec', help='Package written as <user>/<package>/<version>')
    parser.add_argument('--requirements', help='TODO')
    parser.add_argument('--announce', help='markdown announcement to notify watchers of a new release',
                        type=FileType('r'))
    parser.add_argument('--description', help='markdown long description of the package',
                        type=FileType('r'))
    
    
    
    parser.set_defaults(main=main)


'''
Add a new package to your account.
'''
from binstar_client.utils import get_binstar, parse_specs
from argparse import FileType
from os.path import basename
import tarfile
import json
from warnings import warn

def detect_conda_attrs(file):
    try:
        tar = tarfile.open(file)
        obj = tar.extractfile('info/index.json')
        return json.loads(obj.read())
    except:
        warn('Trouble opening conda package')
        return {}

detectors = {'conda':detect_conda_attrs}

def detect(binstar, user, package, file):
    
    package_type = binstar.package(user, package).get('package_type')
    func = detectors.get(package_type, lambda file:{})
    attrs = func(file)
    return attrs

def main(args):
    
    binstar = get_binstar()
    spec = args.spec
    if args.action == 'upload':
        for file in args.files:
            attrs = detect(binstar, spec.user, spec.package, file)
            with open(file) as fd:
                print 'Uploading %s ... ' % file
                binstar.upload(spec.user, spec.package, spec.version, basename(file), fd, args.description, **attrs)
        print '... done'
    elif args.action == 'download':
        fd = binstar.download(spec.user, spec.package, spec.version, spec.basename)
        
        if args.files:
            fname = args.files[0]
        data = fd.read()
        with open(fname, 'w') as fdout:
            fdout.write(data)
    elif args.action == 'list':
        release = binstar.release(spec.user, spec.package, spec.version)
        for dist in release.get('distributions',[]):
            print '%(basename)s id: %(_id)s' % dist 
            for key_value in dist['attrs'].items():
                print '  + %s: %r' % key_value
    elif args.action == 'remove':
        print spec.user, spec.package, spec.version, spec.basename, spec.attrs
        print binstar.remove_dist(spec.user, spec.package, spec.version, spec.basename, spec.attrs)
    else:
        raise NotImplementedError(args.action)
    

def add_parser(subparsers):
    
    parser = subparsers.add_parser('dist',
                                      help='Add a release',
                                      description=__doc__)
    
    parser.add_argument('action', help='Adde remove or update an existing release',
                        choices=['upload', 'download', 'remove', 'list'])
    parser.add_argument('spec', help='Package written as <user>/<package>/<version>', type=parse_specs)
    parser.add_argument('files', nargs='*', help='Distributions to upload', default=[])
    parser.add_argument('-d','--description', help='description of the file(s)')
    
    parser.set_defaults(main=main)


'''
Created on Jun 6, 2013

@author: sean
'''
'''
Created on Jun 6, 2013

@author: sean
'''

from io import BytesIO
from requests.packages.urllib3.filepost import choose_boundary, iter_fields, \
    get_content_type
from requests.packages.urllib3.packages import six
from requests.packages.urllib3.packages.six import b
import codecs
import requests

encoder = codecs.lookup('utf-8')[0]

def writer(lst):
    encoder()
    pass

def encode_multipart_formdata_stream(fields, boundary=None):
    """
    Encode a dictionary of ``fields`` using the multipart/form-data MIME format.

    :param fields:
        Dictionary of fields or list of (key, value) or (key, value, MIME type)
        field tuples.  The key is treated as the field name, and the value as
        the body of the form-data bytes. If the value is a tuple of two
        elements, then the first element is treated as the filename of the
        form-data section and a suitable MIME type is guessed based on the
        filename. If the value is a tuple of three elements, then the third
        element is treated as an explicit MIME type of the form-data section.

        Field names and filenames must be unicode.

    :param boundary:
        If not specified, then a random boundary will be generated using
        :func:`mimetools.choose_boundary`.
    """
    body = []
    def body_write(item):
        if isinstance(item, basestring):
            item = BytesIO(item)
        body.append(item)
        
    body_write_encode = lambda item: body.append(BytesIO(item.encode('utf-8')))
    
    if boundary is None:
        boundary = choose_boundary()

    for fieldname, value in iter_fields(fields):
        body_write(b('--%s\r\n' % (boundary)))

        if isinstance(value, tuple):
            if len(value) == 3:
                filename, data, content_type = value
            else:
                filename, data = value
                content_type = get_content_type(filename)
            body_write_encode('Content-Disposition: form-data; name="%s"; '
                               'filename="%s"\r\n' % (fieldname, filename))
            body_write(b('Content-Type: %s\r\n\r\n' % 
                       (content_type,)))
        else:
            data = value
            body_write_encode('Content-Disposition: form-data; name="%s"\r\n'
                               % (fieldname))
            body_write(b'\r\n')

        if isinstance(data, (int, long)):
            data = str(data)  # Backwards compatibility

        if isinstance(data, six.text_type):
            body_write_encode(data)
        else:
            body_write(data)

        body_write(b'\r\n')

    body_write(b('--%s--\r\n' % (boundary)))

    content_type = b('multipart/form-data; boundary=%s' % boundary)

    return body, content_type

class MultiPartIO(object):
    
    def __init__(self, body, callback=None):
        self.to_read = body
        self.have_read = []
        self._total = 0
        self.callback = callback
        
    def read(self, n=-1):
        if self.callback:
            self.callback(self.tell(), self._total)
            
        if n == -1:
            return ''.join(fd.read() for fd in self.to_read)
            
        if not self.to_read:
            return ''

        while self.to_read: 
            data = self.to_read[0].read(n)
            
            if data:
                return data
            
            fd = self.to_read.pop(0)
            self.have_read.append(fd)
            
        return ''
          
    def tell(self):
        cursor = sum(fd.tell() for fd in self.have_read)
        if self.to_read:
            cursor += self.to_read[0].tell()
        return cursor
            
    def seek(self, pos, mode=0):
        assert pos == 0
        if mode is 0:
            self.to_read = self.have_read + self.to_read
            self.have_read = []
            [fd.seek(pos, mode) for fd in self.to_read]
            self.cursor = 0
            
        elif mode is 2: 
            self.have_read = self.have_read + self.to_read
            self.to_read = []
            [fd.seek(pos, mode) for fd in self.have_read]
            self._total = self.tell()
        
 
def stream_multipart(data, files=None, callback=None):
    from itertools import chain
    if files:
        fields = chain(iter_fields(data), iter_fields(files))
    else:
        fields = data
        
    body, content_type = encode_multipart_formdata_stream(fields)
    data = MultiPartIO(body, callback=callback)
    headers = {'Content-Type':content_type}
    return data, headers


def main():
    def callback(curr, total):
        print '\r%i of %iKb: %.2f%%' % (curr//1024, total//1024, 100.0 * curr / total),
        
    data, headers = stream_multipart({'key1': 'value1', 'key2':'value2'},
                                     files={'file': ('README.md', open('./README.md')),
                                            'file2': ('mongodb-2.4.3-1.tar.bz2', open('mongodb-2.4.3-1.tar.bz2'))},
                                     callback=callback)
    
    requests.post('http://localhost:3339', data=data, headers=headers)



'''
Created on Apr 29, 2013

@author: sean
'''

from hashlib import md5
from keyring import get_keyring
from os.path import exists, join, dirname, expanduser
import appdirs
import base64
import getpass
import os
import urlparse
import yaml
import sys

class UserError(Exception):
    pass

class PackageSpec(object):
    def __init__(self, user, package, version, basename, attrs):
        self._user = user
        self._package = package
        self._version = version
        self._basename = basename
        self.attrs = attrs
    
    @property
    def user(self):
        if self._user is None:
            raise UserError('user not given')
        return self._user

    @property
    def package(self):
        if self._package is None:
            raise UserError('package not given in spec')
        return self._package

    @property
    def version(self):
        if self._version is None:
            raise UserError('version not given in spec')
        return self._version

    @property
    def basename(self):
        if self._basename is None:
            raise UserError('basename not given in spec')
        return self._basename
        
def parse_specs(spec):
    user = spec
    package = version = basename = None
    attrs = {}
    if '/' in user:
        user, package = user.split('/',1)
    if package and '/' in package:
        package, version = package.split('/', 1)
        
    if version and '/' in version:
        version, basename = version.split('/', 1)
    
    if basename and '?' in basename:
        basename, qsl = basename.rsplit('?', 1)
        attrs = dict(urlparse.parse_qsl(qsl))
    
    return PackageSpec(user, package, version, basename, attrs)

def get_binstar():
    from binstar_client import Binstar
    
    kr = get_keyring()
    token = kr.get_password('binstar-token', getpass.getuser())
    url = get_config().get('url', 'https://api.binstar.org')
    
    return Binstar(token, domain=url,)

def load_config(config_file):
    if exists(config_file):
        with open(config_file) as fd:
            data = yaml.load(fd)
            if data: 
                return data
    
    return {}

SITE_CONFIG = join(appdirs.site_data_dir('binstar', 'ContinuumIO'), 'config.yaml')
USER_CONFIG = join(appdirs.user_data_dir('binstar', 'ContinuumIO'), 'config.yaml')
def get_config(user=True, site=True):
    
    config = {}
    if site:
        config.update(load_config(SITE_CONFIG))
    if user:
        config.update(load_config(USER_CONFIG))
        
    return config
    
def set_config(data, user=True):
    
    config_file = USER_CONFIG if user else SITE_CONFIG
    
    data_dir = dirname(config_file)
    if not exists(data_dir):
        os.makedirs(data_dir)
    
    with open(config_file, 'w') as fd:
        yaml.dump(data, fd)



def compute_hash(fp, buf_size=8192, size=None, hash_algorithm=md5):
    hash_obj = hash_algorithm()
    spos = fp.tell()
    if size and size < buf_size:
        s = fp.read(size)
    else:
        s = fp.read(buf_size)
    while s:
        hash_obj.update(s)
        if size:
            size -= len(s)
            if size <= 0:
                break
        if size and size < buf_size:
            s = fp.read(size)
        else:
            s = fp.read(buf_size)
    hex_digest = hash_obj.hexdigest()
    base64_digest = base64.encodestring(hash_obj.digest())
    if base64_digest[-1] == '\n':
        base64_digest = base64_digest[0:-1]
    # data_size based on bytes read.
    data_size = fp.tell() - spos
    fp.seek(spos)
    return (hex_digest, base64_digest, data_size)


class upload_in_chunks(object):
    def __init__(self, fd, chunksize=1 << 13):
        self.fd = fd
        self.chunksize = chunksize
        self.totalsize = os.fstat(fd.fileno()).st_size
        self.readsofar = 0

    def __iter__(self):
        print 'Progress:'
        while True:
            data = self.fd.read(self.chunksize)
            if not data:
                sys.stderr.write("\n")
                break
            self.readsofar += len(data)
            percent = self.readsofar * 1e2 / self.totalsize
            sys.stderr.write("\r{percent:3.0f}%".format(percent=percent))
            yield data

    def __len__(self):
        return self.totalsize


def upload_with_progress(fd):
    it = upload_in_chunks(fd)
    IterableToFileAdapter(it)
    
class IterableToFileAdapter(object):
    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.length = len(iterable)

    def read(self, size=-1): # TBD: add buffer for `len(data) > size` case
        return next(self.iterator, b'')

    def __len__(self):
        return self.length

